@@ -30,19 +30,24 @@
  */
 
 #include "brw_vec4.h"
+#include "brw_cfg.h"
 extern "C" {
 #include "main/macros.h"
 }
 
 namespace brw {
 
+struct copy_entry {
+   src_reg *value[4];
+   int saturatemask;
+};
+
 static bool
 is_direct_copy(vec4_instruction *inst)
 {
    return (inst->opcode == BRW_OPCODE_MOV &&
 	   !inst->predicate &&
 	   inst->dst.file == GRF &&
-	   !inst->saturate &&
 	   !inst->dst.reladdr &&
 	   !inst->src[0].reladdr &&
 	   inst->dst.type == inst->src[0].type);
@@ -73,16 +78,17 @@ is_channel_updated(vec4_instruction *inst, src_reg *values[4], int ch)
 }
 
 static bool
-try_constant_propagation(vec4_instruction *inst, int arg, src_reg *values[4])
+try_constant_propagate(struct brw_context *brw, vec4_instruction *inst,
+                       int arg, struct copy_entry *entry)
 {
    /* For constant propagation, we only handle the same constant
     * across all 4 channels.  Some day, we should handle the 8-bit
     * float vector format, which would let us constant propagate
     * vectors better.
     */
-   src_reg value = *values[0];
+   src_reg value = *entry->value[0];
    for (int i = 1; i < 4; i++) {
-      if (!value.equals(values[i]))
+      if (!value.equals(*entry->value[i]))
 	 return false;
    }
 
@@ -91,18 +97,18 @@ try_constant_propagation(vec4_instruction *inst, int arg, src_reg *values[4])
 
    if (inst->src[arg].abs) {
       if (value.type == BRW_REGISTER_TYPE_F) {
-	 value.imm.f = fabs(value.imm.f);
+	 value.fixed_hw_reg.dw1.f = fabs(value.fixed_hw_reg.dw1.f);
       } else if (value.type == BRW_REGISTER_TYPE_D) {
-	 if (value.imm.i < 0)
-	    value.imm.i = -value.imm.i;
+	 if (value.fixed_hw_reg.dw1.d < 0)
+	    value.fixed_hw_reg.dw1.d = -value.fixed_hw_reg.dw1.d;
       }
    }
 
    if (inst->src[arg].negate) {
       if (value.type == BRW_REGISTER_TYPE_F)
-	 value.imm.f = -value.imm.f;
+	 value.fixed_hw_reg.dw1.f = -value.fixed_hw_reg.dw1.f;
       else
-	 value.imm.u = -value.imm.u;
+	 value.fixed_hw_reg.dw1.ud = -value.fixed_hw_reg.dw1.ud;
    }
 
    switch (inst->opcode) {
@@ -110,6 +116,16 @@ try_constant_propagation(vec4_instruction *inst, int arg, src_reg *values[4])
       inst->src[arg] = value;
       return true;
 
+   case SHADER_OPCODE_POW:
+   case SHADER_OPCODE_INT_QUOTIENT:
+   case SHADER_OPCODE_INT_REMAINDER:
+      if (brw->gen < 8)
+         break;
+      /* fallthrough */
+   case BRW_OPCODE_DP2:
+   case BRW_OPCODE_DP3:
+   case BRW_OPCODE_DP4:
+   case BRW_OPCODE_DPH:
    case BRW_OPCODE_BFI1:
    case BRW_OPCODE_ASR:
    case BRW_OPCODE_SHL:
@@ -151,10 +167,10 @@ try_constant_propagation(vec4_instruction *inst, int arg, src_reg *values[4])
 	 inst->src[arg] = value;
 	 return true;
       } else if (arg == 0 && inst->src[1].file != IMM) {
-	 uint32_t new_cmod;
+	 enum brw_conditional_mod new_cmod;
 
 	 new_cmod = brw_swap_cmod(inst->conditional_mod);
-	 if (new_cmod != ~0u) {
+	 if (new_cmod != BRW_CONDITIONAL_NONE) {
 	    /* Fit this constant in by swapping the operands and
 	     * flipping the test.
 	     */
@@ -191,24 +207,33 @@ try_constant_propagation(vec4_instruction *inst, int arg, src_reg *values[4])
    return false;
 }
 
-bool
-vec4_visitor::try_copy_propagation(vec4_instruction *inst, int arg,
-                                   src_reg *values[4])
+static bool
+is_logic_op(enum opcode opcode)
+{
+   return (opcode == BRW_OPCODE_AND ||
+           opcode == BRW_OPCODE_OR  ||
+           opcode == BRW_OPCODE_XOR ||
+           opcode == BRW_OPCODE_NOT);
+}
+
+static bool
+try_copy_propagate(struct brw_context *brw, vec4_instruction *inst,
+                   int arg, struct copy_entry *entry, int reg)
 {
    /* For constant propagation, we only handle the same constant
     * across all 4 channels.  Some day, we should handle the 8-bit
     * float vector format, which would let us constant propagate
     * vectors better.
     */
-   src_reg value = *values[0];
+   src_reg value = *entry->value[0];
    for (int i = 1; i < 4; i++) {
       /* This is equals() except we don't care about the swizzle. */
-      if (value.file != values[i]->file ||
-	  value.reg != values[i]->reg ||
-	  value.reg_offset != values[i]->reg_offset ||
-	  value.type != values[i]->type ||
-	  value.negate != values[i]->negate ||
-	  value.abs != values[i]->abs) {
+      if (value.file != entry->value[i]->file ||
+	  value.reg != entry->value[i]->reg ||
+	  value.reg_offset != entry->value[i]->reg_offset ||
+	  value.type != entry->value[i]->type ||
+	  value.negate != entry->value[i]->negate ||
+	  value.abs != entry->value[i]->abs) {
 	 return false;
       }
    }
@@ -219,7 +244,7 @@ vec4_visitor::try_copy_propagation(vec4_instruction *inst, int arg,
     */
    int s[4];
    for (int i = 0; i < 4; i++) {
-      s[i] = BRW_GET_SWZ(values[i]->swizzle,
+      s[i] = BRW_GET_SWZ(entry->value[i]->swizzle,
 			 BRW_GET_SWZ(inst->src[arg].swizzle, i));
    }
    value.swizzle = BRW_SWIZZLE4(s[0], s[1], s[2], s[3]);
@@ -229,6 +254,11 @@ vec4_visitor::try_copy_propagation(vec4_instruction *inst, int arg,
        value.file != ATTR)
       return false;
 
+   if (brw->gen >= 8 && (value.negate || value.abs) &&
+       is_logic_op(inst->opcode)) {
+      return false;
+   }
+
    if (inst->src[arg].abs) {
       value.negate = false;
       value.abs = true;
@@ -242,12 +272,16 @@ vec4_visitor::try_copy_propagation(vec4_instruction *inst, int arg,
     * instructions.
     */
    if ((has_source_modifiers || value.file == UNIFORM ||
-        value.swizzle != BRW_SWIZZLE_XYZW) && !can_do_source_mods(inst))
+        value.swizzle != BRW_SWIZZLE_XYZW) && !inst->can_do_source_mods(brw))
       return false;
 
    if (has_source_modifiers && value.type != inst->src[arg].type)
       return false;
 
+   if (has_source_modifiers &&
+       inst->opcode == SHADER_OPCODE_GEN4_SCRATCH_WRITE)
+      return false;
+
    bool is_3src_inst = (inst->opcode == BRW_OPCODE_LRP ||
                         inst->opcode == BRW_OPCODE_MAD ||
                         inst->opcode == BRW_OPCODE_BFE ||
@@ -268,9 +302,28 @@ vec4_visitor::try_copy_propagation(vec4_instruction *inst, int arg,
       return false;
 
    /* Don't report progress if this is a noop. */
-   if (value.equals(&inst->src[arg]))
+   if (value.equals(inst->src[arg]))
       return false;
 
+   /* Limit saturate propagation only to SEL with src1 bounded within 1.0 and 1.0
+    * otherwise, skip copy propagate altogether
+    */
+   if (entry->saturatemask & (1 << arg)) {
+      switch(inst->opcode) {
+      case BRW_OPCODE_SEL:
+         if (inst->src[1].file != IMM ||
+             inst->src[1].fixed_hw_reg.dw1.f < 0.0 ||
+             inst->src[1].fixed_hw_reg.dw1.f > 1.0) {
+            return false;
+         }
+         if (!inst->saturate)
+            inst->saturate = true;
+         break;
+      default:
+         return false;
+      }
+   }
+
    value.type = inst->src[arg].type;
    inst->src[arg] = value;
    return true;
@@ -280,13 +333,11 @@ bool
 vec4_visitor::opt_copy_propagation()
 {
    bool progress = false;
-   src_reg *cur_value[virtual_grf_reg_count][4];
+   struct copy_entry entries[virtual_grf_reg_count];
 
-   memset(&cur_value, 0, sizeof(cur_value));
-
-   foreach_list(node, &this->instructions) {
-      vec4_instruction *inst = (vec4_instruction *)node;
+   memset(&entries, 0, sizeof(entries));
 
+   foreach_block_and_inst(block, vec4_instruction, inst, cfg) {
       /* This pass only works on basic blocks.  If there's flow
        * control, throw out all our information and start from
        * scratch.
@@ -295,7 +346,7 @@ vec4_visitor::opt_copy_propagation()
        * src/glsl/opt_copy_propagation.cpp to track available copies.
        */
       if (!is_dominated_by_previous_instruction(inst)) {
-	 memset(cur_value, 0, sizeof(cur_value));
+	 memset(&entries, 0, sizeof(entries));
 	 continue;
       }
 
@@ -316,31 +367,38 @@ vec4_visitor::opt_copy_propagation()
 
 	 /* Find the regs that each swizzle component came from.
 	  */
-	 src_reg *values[4];
+         struct copy_entry entry;
+         memset(&entry, 0, sizeof(copy_entry));
 	 int c;
 	 for (c = 0; c < 4; c++) {
-	    values[c] = cur_value[reg][BRW_GET_SWZ(inst->src[i].swizzle, c)];
+            int channel = BRW_GET_SWZ(inst->src[i].swizzle, c);
+            entry.value[c] = entries[reg].value[channel];
 
 	    /* If there's no available copy for this channel, bail.
 	     * We could be more aggressive here -- some channels might
 	     * not get used based on the destination writemask.
 	     */
-	    if (!values[c])
+	    if (!entry.value[c])
 	       break;
 
+            entry.saturatemask |=
+               (entries[reg].saturatemask & (1 << channel) ? 1 : 0) << c;
+
 	    /* We'll only be able to copy propagate if the sources are
 	     * all from the same file -- there's no ability to swizzle
 	     * 0 or 1 constants in with source registers like in i915.
 	     */
-	    if (c > 0 && values[c - 1]->file != values[c]->file)
+	    if (c > 0 && entry.value[c - 1]->file != entry.value[c]->file)
 	       break;
 	 }
 
 	 if (c != 4)
 	    continue;
 
-	 if (try_constant_propagation(inst, i, values) ||
-	     try_copy_propagation(inst, i, values))
+	 if (try_constant_propagate(brw, inst, i, &entry))
+            progress = true;
+
+	 if (try_copy_propagate(brw, inst, i, &entry, reg))
 	    progress = true;
       }
 
@@ -354,9 +412,11 @@ vec4_visitor::opt_copy_propagation()
 	  * the new value, so clear it.
 	  */
 	 bool direct_copy = is_direct_copy(inst);
+	 entries[reg].saturatemask = 0x0;
 	 for (int i = 0; i < 4; i++) {
 	    if (inst->dst.writemask & (1 << i)) {
-	       cur_value[reg][i] = direct_copy ? &inst->src[0] : NULL;
+               entries[reg].value[i] = (!inst->saturate && direct_copy) ? &inst->src[0] : NULL;
+               entries[reg].saturatemask |= (((inst->saturate && direct_copy) ? 1 : 0) << i);
 	    }
 	 }
 
@@ -364,13 +424,14 @@ vec4_visitor::opt_copy_propagation()
 	  * our destination's updated channels, as the two are no longer equal.
 	  */
 	 if (inst->dst.reladdr)
-	    memset(cur_value, 0, sizeof(cur_value));
+	    memset(&entries, 0, sizeof(entries));
 	 else {
 	    for (int i = 0; i < virtual_grf_reg_count; i++) {
 	       for (int j = 0; j < 4; j++) {
-		  if (is_channel_updated(inst, cur_value[i], j)){
-		     cur_value[i][j] = NULL;
-		  }
+		  if (is_channel_updated(inst, entries[i].value, j)){
+		     entries[i].value[j] = NULL;
+		     entries[i].saturatemask &= ~(1 << j);
+                  }
 	       }
 	    }
 	 }
