@@ -192,9 +192,9 @@ get_texture_format_swizzle(const struct st_texture_object *stObj)
    return swizzle_swizzle(stObj->base._Swizzle, tex_swizzle);
 }
 
-                            
+
 /**
- * Return TRUE if the texture's sampler view swizzle is equal to
+ * Return TRUE if the texture's sampler view swizzle is not equal to
  * the texture's swizzle.
  *
  * \param stObj  the st texture object,
@@ -212,6 +212,23 @@ check_sampler_swizzle(const struct st_texture_object *stObj,
 }
 
 
+static unsigned last_level(struct st_texture_object *stObj)
+{
+   unsigned ret = MIN2(stObj->base.MinLevel + stObj->base._MaxLevel,
+                       stObj->pt->last_level);
+   if (stObj->base.Immutable)
+      ret = MIN2(ret, stObj->base.MinLevel + stObj->base.NumLevels - 1);
+   return ret;
+}
+
+static unsigned last_layer(struct st_texture_object *stObj)
+{
+   if (stObj->base.Immutable && stObj->pt->array_size > 1)
+      return MIN2(stObj->base.MinLayer + stObj->base.NumLayers - 1,
+                  stObj->pt->array_size - 1);
+   return stObj->pt->array_size - 1;
+}
+
 static struct pipe_sampler_view *
 st_create_texture_sampler_view_from_stobj(struct pipe_context *pipe,
 					  struct st_texture_object *stObj,
@@ -243,7 +260,13 @@ st_create_texture_sampler_view_from_stobj(struct pipe_context *pipe,
       templ.u.buf.first_element = f;
       templ.u.buf.last_element  = f + (n - 1);
    } else {
-      templ.u.tex.first_level = stObj->base.BaseLevel;
+      templ.u.tex.first_level = stObj->base.MinLevel + stObj->base.BaseLevel;
+      templ.u.tex.last_level = last_level(stObj);
+      assert(templ.u.tex.first_level <= templ.u.tex.last_level);
+      templ.u.tex.first_layer = stObj->base.MinLayer;
+      templ.u.tex.last_layer = last_layer(stObj);
+      assert(templ.u.tex.first_layer <= templ.u.tex.last_layer);
+      templ.target = gl_target_to_pipe(stObj->base.Target);
    }
 
    if (swizzle != SWIZZLE_NOOP) {
@@ -271,11 +294,19 @@ st_get_texture_sampler_view_from_stobj(struct st_context *st,
 
    sv = st_texture_get_sampler_view(st, stObj);
 
+   if (stObj->base.StencilSampling &&
+       util_format_is_depth_and_stencil(format))
+      format = util_format_stencil_only(format);
+
    /* if sampler view has changed dereference it */
    if (*sv) {
       if (check_sampler_swizzle(stObj, *sv) ||
 	  (format != (*sv)->format) ||
-	  stObj->base.BaseLevel != (*sv)->u.tex.first_level) {
+          gl_target_to_pipe(stObj->base.Target) != (*sv)->target ||
+          stObj->base.MinLevel + stObj->base.BaseLevel != (*sv)->u.tex.first_level ||
+          last_level(stObj) != (*sv)->u.tex.last_level ||
+          stObj->base.MinLayer != (*sv)->u.tex.first_layer ||
+          last_layer(stObj) != (*sv)->u.tex.last_layer) {
 	 pipe_sampler_view_reference(sv, NULL);
       }
    }
@@ -325,7 +356,7 @@ update_single_texture(struct st_context *st,
    /* Determine the format of the texture sampler view */
    if (texObj->Target == GL_TEXTURE_BUFFER) {
       view_format =
-         st_mesa_format_to_pipe_format(stObj->base._BufferObjectFormat);
+         st_mesa_format_to_pipe_format(st, stObj->base._BufferObjectFormat);
    }
    else {
       view_format =
@@ -354,7 +385,7 @@ update_textures(struct st_context *st,
 {
    const GLuint old_max = *num_textures;
    GLbitfield samplers_used = prog->SamplersUsed;
-   GLuint unit, new_count;
+   GLuint unit;
 
    if (samplers_used == 0x0 && old_max == 0)
       return;
@@ -383,16 +414,9 @@ update_textures(struct st_context *st,
       pipe_sampler_view_reference(&(sampler_views[unit]), sampler_view);
    }
 
-   /* Ex: if old_max = 3 and *num_textures = 1, we need to pass an
-    * array of views={X, NULL, NULL} to unref the old texture views
-    * at positions [1] and [2].
-    */
-   new_count = MAX2(*num_textures, old_max);
-   assert(new_count <= max_units);
-
    cso_set_sampler_views(st->cso_context,
                          shader_stage,
-                         new_count,
+                         *num_textures,
                          sampler_views);
 }
 
@@ -437,7 +461,7 @@ update_geometry_textures(struct st_context *st)
       update_textures(st,
                       PIPE_SHADER_GEOMETRY,
                       &ctx->GeometryProgram._Current->Base,
-                      ctx->Const.Program[MESA_SHADER_FRAGMENT].MaxTextureImageUnits,
+                      ctx->Const.Program[MESA_SHADER_GEOMETRY].MaxTextureImageUnits,
                       st->state.sampler_views[PIPE_SHADER_GEOMETRY],
                       &st->state.num_sampler_views[PIPE_SHADER_GEOMETRY]);
    }
